#include <Arduino.h>
#include <stdio.h>
#include <string.h>
#include "cDS345.h"

// Constructors, Destructors //////////////////////////////////////////////////
cDS345::cDS345(/* args */) :
    m_fFrequency(1000.0f),
    m_fAmplitude(0.0f),
    m_fOffset(0.0f),
    m_Channel(CH1),
    m_WaveType(SINE),
    m_fLoadFactor(1.0f)
{
    memset(m_sResponse, 0, RESPONSE_LENGTH);
    memset(m_buf, 0, BUF_SIZE);
}

cDS345::~cDS345()
{
}


/**
 * @brief Send a command to our generator via Serial2
 * 
 * @param cmd: Command string as char[]
 */
void cDS345::sendCommand(const char* cmd)
{
    Serial2.println(cmd);
    delay(40);     // Give DS345 time to digest

    ESP_LOGI("", "SENT TO DS345: \"%s\"\n", cmd);
}


/**
 * @brief Receive a  response from our generator via Serial2
 * 
 * Timeout after 1 second.
 * 
 * @return  A pointer to the response string
 *          NULL on timeout reading from generator
 */
char* cDS345::readResponse(void)
{
    static const unsigned long TIMEOUT = 1000;
    unsigned long startTime = millis();
    while (Serial2.available() == 0) {
        if ((millis() - startTime) > TIMEOUT) {
            memset(m_sResponse, 0, RESPONSE_LENGTH);
            ESP_LOGW("", "Timeout reading from DS345");
            return NULL;
        }
    }
    size_t len = Serial2.read(m_sResponse, RESPONSE_LENGTH);
    m_sResponse[len] = 0;

    ESP_LOGI("", "REC'D FROM DS345: \"%s\"\n", m_sResponse);

    return m_sResponse;
}


/**
 * @brief convert a waveform code to its text equivalent
 * 
 * The text equivalent conforms to Siglent SDG programming guide
 * @param   wave: Waveform code as defined in enum wave_t
 * @param   text: an external character buffer of min. char[8]
 */
void cDS345::wave2text(wave_t wave, char** text)
{
    switch (wave) {
    case SINE:
        strcpy(*text, "SINE");
        break;
    case SQUARE:
        strcpy(*text, "SQUARE");
        break;
    case TRIANGLE:
        // Triangle exists in DS345 but not in Siglent's SDG series
        // so fall through to RAMP
    case RAMP:
        strcpy(*text, "RAMP");   // No closer text available
        break;
    case NOISE:
        strcpy(*text, "NOISE");
        break;
    case ARBITRARY:
        strcpy(*text, "ARB");
        break;
    default:
        strcpy(*text, "unknown");
        break;
    }
}


/**
 * @brief Reset DS345 to a "harmless" state, called at startup
 */
void cDS345::reset(void)
{
    sendCommand("*CLS");
    sendCommand("FUNC 0");
    sendCommand("FREQ 1000");
    sendCommand("AMPL 0VP");
    sendCommand("OFFS 0");
    sendCommand("PCLR");
    sendCommand("INVT 0");
    sendCommand("MENA 0");
}


/**
 * @brief Set the output wave type
 * 
 * This should only be used with SINE wavetype in the context of Bode-plot.
 * Coincidentally code 0 represents sine wave for Siglent as well as for DS345.
 * 
 * @param type: The wavetype code e.g. 0 = SINE
 */
void cDS345::setWave(wave_t type)
{
    m_WaveType = type;
    snprintf(m_buf, BUF_SIZE, "FUNC %d", type);
    sendCommand(m_buf);
}


/**
 * @brief Retrieve DS345's output wave type
 * 
 * First read the wave type from DS345, store it locally and return its code
 * to the caller
 * 
 * @return  The wave type code e.g. 0 = SINE
 */
wave_t cDS345::getWave(void)
{
    readWave();
    return m_WaveType;
}


/**
 * @brief Set the output frequency in Hertz
 * 
 * Store the frequency locally and send the corresponding command to DS345
 * 
 * @param freq  The frequency in Hz
 */
void cDS345::setFrequency(double freq)
{
    m_fFrequency = freq;
    snprintf(m_buf, BUF_SIZE, "FREQ %.3f", freq);
    sendCommand(m_buf);
}


/**
 * @brief Retrieve DS345's frquency setting
 * 
 * First read the frequency from DS345, store it locally and return it to
 * the caller.
 * 
 * @return the frequency in Hz
 */
double cDS345::getFrequency(void)
{
    readFrequency();
    return m_fFrequency;
}


/**
 * @brief Set the generator's output amplitude in Vpeak-peak
 * 
 * The output amplitude depends on the generator's load impedance.
 * The DS345 generator has 50 ohm source impedance. The output voltage in
 * terms of amplitude and offset are only correct for a 50 ohms load
 * impedance. Otherwise the values for amplitude and offset must be corrected.
 * This is done via function load(), setting m_fLoadFactor.
 * So: <DS345 internal setting> = m_fAmplitude / m_fLoadFactor
 *
 * @param ampl The amplitude in Vpp
 */
void cDS345::setAmplitude(float ampl)
{
    m_fAmplitude = ampl;
    snprintf(m_buf, BUF_SIZE, "AMPL %.3fVP", m_fAmplitude / m_fLoadFactor);
    sendCommand(m_buf);
}


/**
 * @brief Retrieve DS345's output amplitude, corrected by it's load factor
 * 
 * The function updates the offset voltage setting from DS345, applying
 * the load impedance correction factor and stores the result locally.
 *
 * @see setAmplitude()
 * @see setLoadFactor()
 * 
 * @return The effective output amplitude in Vpeak-peak, taking into account
 *          the load impedance.
 */
float cDS345::getAmplitude(void)
{
    readAmplitude();
    return m_fAmplitude;
}


/**
 * @brief Set the generator's offset voltage
 * 
 * The effective offset depends on the generator's load impedance.
 * The DS345 generator has 50 ohm source impedance. The output voltage in
 * terms of amplitude and offset are only correct for a 50 ohms load
 * impedance. Otherwise the values for amplitude and offset must be corrected.
 * This is done via function load() setting m_fLoadFactor
 * So: <DS345 internal setting> = m_fOffset / m_fLoadFactor
 * 
 * @param offs  The effective offset voltage in volts
 */
void cDS345::setOffset(float offs)
{
    m_fOffset = offs;
    snprintf(m_buf, BUF_SIZE, "OFFS %.3f", m_fOffset / m_fLoadFactor);
    sendCommand(m_buf);
}


/**
 * @brief Return the generator's effective offset voltage
 * 
 * The function updates the offset voltage setting from DS345, applying
 * the load impedance correction factor and stores the result locally.
 *
 * @see setOffset()
 * @see setLoadFactor()
 * 
 * @return The effective generator's offset voltage in volts taking into 
 *          account the load impedance
 */
float cDS345::getOffset(void)
{
    readOffset();
    return m_fOffset;
}


/**
 * @brief Set an output voltage correction factor based on load resistance
 * 
 * The DS345 generator has 50 ohm output impedance. The output voltage in
 * terms of amplitude and offset are only correct for a 50 ohms load
 * impedance. Otherwise the values for amplitude and offset must be corrected.
 * 
 * @see setAmplitude()
 * @see setOffset()
 * 
 * @param load  Load impedance in ohms
 */
void cDS345::setLoadFactor(unsigned int load)
{
    m_fLoadFactor = (2.0f * (double)load) / (50.0f + (double)load);
    // Update DS345 internal setting using the new load factor
    setAmplitude(m_fAmplitude);     // Re-use the current amplitude value
    setOffset(m_fOffset);           // Re-use the current offset value
}


/**
 * @brief Get all relevant parameters from DS345
 * 
 * The function retrieves values for waveform, frequency, amplitude and offset
 * voltage from DS345 and formats them in a string digestible by SDS1004X-E.
 * 
 * @param paramString: A pointer to a zero-terminated character string residing
 *          as a member variable in the ds345 object where the result for the 
 *          scope is stored. The caller can access the result using that pointer.
 * @return The length of the resulting charater string
 */
size_t cDS345::getAllParams(char** paramString)
{
    readWave();
    char waveBuffer[10];
    char* pBuf = waveBuffer;
    wave2text(m_WaveType, &pBuf);
    readFrequency();
    readAmplitude();
    readOffset();

    snprintf(m_buf, BUF_SIZE, "C1:BSWV WVTP,%s,FRQ,%.3fHZ,AMP,%.3fV,OFST,%.3fV,PHSE,0",
        waveBuffer, m_fFrequency, m_fAmplitude, m_fOffset);
    *paramString = m_buf;
    return strlen(m_buf);
}


/**
 * @brief Helper function to read a parameter from DS345
 * 
 * @param param: A character string containing the command for DS345
 * 
 * @return the response string of DS345 or NULL if the request times out
 */
char* cDS345::readParam(const char* param)
{
    sendCommand(param);
    return readResponse();
}


/**
 * @brief Get the generator's output waveform code and store ot locally
 * 
 * By convention this should always be code 0 (SINE)
 */
void cDS345::readWave(void)
{
    char* pFunc = readParam("FUNC?");
    if (pFunc != NULL) {
        m_WaveType = (wave_t)atoi(pFunc);
    } else {
        // Leave m_WaveType unchanged on timeout
    }
}


/**
 * @brief Get the generator's frequency in Hz and store it locally
 */
void cDS345::readFrequency(void)
{
    char* pFreq = readParam("FREQ?");
    if (pFreq != NULL) {
        m_fFrequency = atof(pFreq);
    } else {
        // Leave m_fFrequency unchanged on timeout
    }
}


/**
 * @brief Get the generator's output amplitude and store it locally
 * 
 * Caution: The effective output amplitude is load-dependent! It is equal to
 * the generator's internal settings only if the load impedance = 50 Ohms.
 * For infinite load impedance its value is double the DS345's programming.
 * The (private) member variable m_fLoadFactor takes care of that:
 *    <DS345 internal setting> * m_fLoadFactor = <effective value>
 * 
 * The function reads the programmed amplitude, multiplies it with m_fLoadFactor
 * and stores the result in m_fAmplitude. By convention the amplitude is a
 * paek-peak value.
 */
void cDS345::readAmplitude(void)
{
    char* pAmpl = readParam("AMPL?");
    if (pAmpl != NULL) {
         m_fAmplitude = atof(pAmpl) * m_fLoadFactor;
    } else {
        // Leave m_fAmplitude unchanged on timeout
    }
}


/**
 * @brief Get the generator's offset voltage and store it locally
 * 
 * Caution: The effective offset voltage is load-dependent! It is equal to
 * the generator's internal settings only if the load impedance = 50 Ohms.
 * For infinite load impedance its value is double the DS345's programming.
 * The (private) member variable m_fLoadFactor takes care of that:
 *    <DS345 internal setting> * m_fLoadFactor = <effective value>
 * 
 * The function reads the programmed offset, multiplies it with m_fLoadFactor
 * and stores the result in m_fOffset.
 */
void cDS345::readOffset(void)
{
    char* pOffs = readParam("OFFS?");
    if (pOffs != NULL) {
        m_fOffset = atof(pOffs) * m_fLoadFactor;
    } else {
        // Leave m_fOffset unchanged on timeout
    }
}