/**
 * @file esp32Bode.coo
 * 
 * Interface between a Siglent oscilloscope of the SDS1004X-E series and a
 * Synthesized Function Generator DS345 of Stanford Research Systems.
 * 
 * The program is heavily based on work of "AimcLsc": 
 * https://github.com/AimcLsc/espBode   see also for a discussion (comment #42)
 * https://www.eevblog.com/forum/testgear/siglent-sds1104x-e-and-sds1204x-e-bode-plot-with-non-siglent-awg/
 * 
 * Communication to and from the oscilloscope works over WLAN using the 
 * VXI-11 protocol. Communication to and from the generator uses RS-232 over
 * Serial2.
 * 
 * History:
 * 2021-05-02 mTron47:
 *      * Adaptation to ESP32 with Platformio as IDE
 *      * Replaced files fy6800.cpp and fy6800.h, by new class cDS345 in files
 *        cDS345.cpp and cDS345.h
 *      * Modified function handleWriteMsg() making parseDecimal() redundant
 *      * Corrected bug in sendReadResponse() leading to heap corruption
 *      * Hid WLAN credentials in "secret" file credentials.h
 *      * Added a few niceties like a monitor LED to watch operation or
 *        a greeting message with file name and compilation date for Serial
 */

#include <Arduino.h>
#include <WiFi.h>
#include "credentials.h"
#include "esp_network.h"
#include "esp_config.h"
#include "cDS345.h"

// Global Variables, Objects //////////////////////////////////////////////////
WiFiServer rpc_server(RPC_PORT);
WiFiServer lxi_server(LXI_PORT);

cDS345 ds345;               // Our instance of SRS DS345 generator


// Helper Functions ///////////////////////////////////////////////////////////
/**
 * @brief Send App-name, compile-date & -time to Serial
 * 
 * @param baud: Baud rate for Serial. If undefined it is assumed that Serial
 *              has been intialized externally before calling sayHello()
 */
void sayHello(unsigned long baud = 0)
{
    if (baud > 0) {
        Serial.begin(baud);
    }
    String sHello("\n");
    sHello += _NAME_;
    sHello += "   ";
    sHello += __DATE__;
    sHello += " ";
    sHello += __TIME__;
    Serial.println(sHello);
    for (size_t i = 0; i < (sHello.length() - 1); i++) 
        Serial.print("*");
    Serial.println("");
}


// Arduino setup() Function ///////////////////////////////////////////////////
void setup() {

    sayHello(115200);
    Serial2.begin(19200, SERIAL_8N2);
    ds345.reset();

    pinMode(LED_PIN, OUTPUT);
    digitalWrite(LED_PIN, LOW);

    // We start by connecting to a WiFi network
    Serial.print("Connecting to ");
    Serial.print(mySSID);
    Serial.print(" ");

#if defined(STATIC_IP)
    IPAddress ip(ESP_IP);
    IPAddress mask(ESP_MASK);
    IPAddress gateway(ESP_GW);
    WiFi.config(ip, gateway, mask);
#endif

#if defined(WIFI_MODE_CLIENT)
    WiFi.mode(WIFI_STA);
    WiFi.begin(mySSID, myPASSWORD);
#elif defined(WIFI_MODE_AP)
    WiFi.softAP(WIFI_SSID, WIFI_PSK);
#else
    #error PLEASE SELECT WIFI_MODE_AP OR WIFI_MODE_CLIENT!
#endif

    while (WiFi.status() != WL_CONNECTED) {
        Serial.print(".");
        digitalWrite(LED_PIN, HIGH);
        delay(100);
        digitalWrite(LED_PIN, LOW);
        delay(150);
    }
    Serial.println (" done");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());

    rpc_server.begin();
    lxi_server.begin();
    digitalWrite(LED_PIN, HIGH);
}


// Arduino work loop() ////////////////////////////////////////////////////////
void loop() {
    /*// Start of test sequence ////////////////////////////////////////////////
    char quest[100];
    size_t i = 0;
    char inChar[2];

    while (true) {
        if (Serial.readBytes(inChar, 1) > 0) {
            if (inChar[0] == '\n') {
                quest[i] = 0;
                ds345.sendCommand(quest);
                break;
            }
            quest[i] = inChar[0];
            i++;
        }
    }
    Serial.println(ds345.readResponse());

    char* pAll;
    ds345.getAllParams(&pAll);
    Serial.println(pAll);
    return;
    // End of test sequence ///////////////////////////////////////////////////*/


    WiFiClient rpc_client;
    WiFiClient lxi_client;

    lxi_client.setTimeout(1000);

    do {
        rpc_client = rpc_server.available();
    } while (!rpc_client);
    DEBUG("RPC CONNECTION.");

    digitalWrite(LED_PIN, LOW); // LED OFF during transaction
    handlePacket(rpc_client);
    rpc_client.stop();

    do {
        lxi_client = lxi_server.available();
    } while (!lxi_client);
    DEBUG("LXI CONNECTION.");

    while (true) {
        if (0 != handlePacket(lxi_client)) {
            lxi_client.stop();
            digitalWrite(LED_PIN, HIGH);
            DEBUG("RESTARTING");
            return;
        }
    }
}
